"""PowerGrASP package.

"""

from . import cli
from .routines import compress_by_cc
from .constants import print_config


def run_cli():
    args = cli.parse_args(__doc__)

    if args.show_config:
        print_config()
        exit()
    elif args.infile:
        with open(args.outfile, 'w') as fd:
            for line in compress_by_cc(args.infile, args.recipe):
                fd.write(line + '\n')
    else:
        print('Nothing to do.')


if __name__ == "__main__":
    run_cli()



"""Routines implementing edge filtering on input graphs,
knowing the bounds on the size of the motif to search in it.

All routines follow the following interface:
- arguments are the graph and the bounds
- yield valid edges as pairs of nodes

"""

import networkx as nx


def for_biclique(graph:nx.Graph, lowerbound:int, upperbound:int) -> [(str, str)]:
    """
    Remove any edge implying a node of degree 1, and a node of degree < lowerbound.
    """
    def ok_to_go(edge:tuple) -> bool:
        one, two = edge
        degone, degtwo = map(graph.degree, edge)
        return not any((
            degone == 1 and degtwo < lowerbound,
            degtwo == 1 and degone < lowerbound,
        ))
    for edge in graph.edges:
        if ok_to_go(edge):
            yield edge


def for_star(graph:nx.Graph, lowerbound:int, upperbound:int) -> [(str, str)]:
    """
    """
    def ok_to_go(edge:tuple) -> bool:
        one, two = edge
        degone, degtwo = map(graph.degree, edge)
        return not all((
            degone < lowerbound,
            degtwo < lowerbound,
        ))
    for edge in graph.edges:
        if ok_to_go(edge):
            yield edge


def for_clique(graph:nx.Graph, lowerbound:int, upperbound:int) -> [(str, str)]:
    """
    Remove an edge when one participating node has a clustering coefficient equal to 0.
    """
    clusterings = {}  # node -> clustering coefficient
    def clustering_of(node:str) -> float:
        if node in clusterings:
            return clusterings[node]
        else:
            return clusterings.setdefault(node, nx.clustering(graph, node))
    for edge in graph.edges:
        if all(clustering_of(node) > 0. for node in edge):
            yield edge



"""Definition of high level functions operating the compression.

"""

import csv
from .searchers import CliqueSearcher, BicliqueSearcher, StarSearcher, NonStarBicliqueSearcher
from .utils import get_time
from .graph import Graph
from .recipe import Recipe
from . import constants as const
from .constants import MULTISHOT_MOTIF_SEARCH, BUBBLE_FOR_EACH_STEP, TIMERS, SHOW_STORY, SHOW_DEBUG, STATISTIC_FILE, USE_STAR_MOTIF
from .motif_batch import MotifBatch
from multiprocessing.dummy import Pool as ThreadPool  # dummy here to use the threading backend, not process
from multiprocessing import Pool as ProcessPool


if TIMERS and STATISTIC_FILE:
    # function to fill the file during compression
    def save_stats(*args):
        """Fill statistic file with given data"""
        with open(STATISTIC_FILE, 'a') as fd:
            fd.write(','.join(map(str, args)) + '\n')


def search_best_motifs_sequentially(searchers, step, recipe) -> MotifBatch:
    """Return a MotifBatch instance containing the best motifs
    found by given searchers."""
    score_to_beat = 0
    best_motifs, best_motifs_score = None, 0
    ordered_searchers = const.MOTIF_TYPE_ORDER(searchers)
    for searcher in ordered_searchers:
        motifs = MotifBatch(searcher.search(step, score_to_beat, recipe=recipe))
        if motifs:
            searcher.on_new_found_motif(motifs)
            if motifs.score > best_motifs_score:
                best_motifs, best_motifs_score = motifs, motifs.score
                score_to_beat = best_motifs_score
    return best_motifs

def search_best_motifs_in_parallel(searchers, step, recipe) -> MotifBatch:
    """Return a MotifBatch instance containing the best motifs
    found by given searchers."""
    with ThreadPool(len(searchers)) as pool:
        founds = pool.starmap(search_best_motifs_sequentially, (([s], step, recipe) for s in searchers))
    return max(founds, key=lambda f: 0 if f is None else f.score)

if const.PARALLEL_MOTIF_SEARCH:
    search_best_motifs = search_best_motifs_in_parallel
else:
    search_best_motifs = search_best_motifs_sequentially


def compress(graph:Graph, *, cc_idx=None, recipe:[Recipe]=None) -> [str]:
    """Yield bubble lines found in graph"""
    if TIMERS:
        timer_start = get_time()
        timer_last = timer_start
    if USE_STAR_MOTIF:
        searchers = CliqueSearcher(graph), NonStarBicliqueSearcher(graph), StarSearcher(graph)
    else:
        searchers = CliqueSearcher(graph), BicliqueSearcher(graph)
    if SHOW_STORY:
        print('INFO searchers: ' + ', '.join(s.name for s in searchers))
        print(f"INFO recipe: {recipe}")
    recipe_lines, recipe_line, recipe_completed = iter(recipe or ()), None, False
    step = 0
    complete_compression = False
    while True:
        if recipe_line and recipe_line.isbreakable and not recipe_completed:
            pass  # reuse the same recipe line
        else:  # everything normal
            recipe_line = next(recipe_lines, None)
            recipe_completed = False
        if (SHOW_STORY and recipe_line) or SHOW_DEBUG: print('INFO recipe:', recipe_line)
        step += 1
        try:
            best_motifs = search_best_motifs(searchers, step, recipe=recipe_line)
        except KeyboardInterrupt:
            print('WARNING interrupted search. Graph compression aborted. Output will be written.')
            best_motifs = None
            break
        if best_motifs:  # let's compress it
            step += graph.compress_all(best_motifs.non_overlapping_subset())
            for searcher in searchers:
                searcher.on_new_compressed_motif(best_motifs)
            if BUBBLE_FOR_EACH_STEP:
                graph.output('out/out_k{}_s{}.bbl'.format(step, best_motif.score))
            if SHOW_STORY:
                print('INFO {} {} motif of score {} compressed'.format(best_motifs.count, best_motifs.name, best_motifs.score))
            if TIMERS:
                now = get_time()
                timers = round(now - timer_start, 2), round(now - timer_last, 2)
                if SHOW_STORY:
                    print("TIMER since start: {}s\t\tsince last motif: {}s"
                          "".format(*timers))
                timer_last = now
                for searcher in searchers:  # timer per motif search
                    timers = timers + ('{}:{}'.format(searcher.name, round(searcher.last_search_time, 2)),)
            if STATISTIC_FILE:
                bounds = [
                    '{}:[{};{}]'.format(searcher.name, searcher.lowerbound, searcher.upperbound)
                    for searcher in searchers
                ]
                if not TIMERS:
                    timers = 'none', 'none'
                save_stats(cc_idx, *timers, best_motifs.name, best_motifs.score, *bounds)
        else:
            if recipe_line:  # the recipe failed, or is optional
                recipe_completed = True
                if recipe_line.isbreakable:
                    if SHOW_DEBUG:
                        print(f"DEBUG breakable recipe {recipe_line} exhausted.")
                elif recipe_line.isrequired:
                    raise recipe_line.RecipeError(f"Recipe {recipe_line} failed, but was necessary.")
                else:  # it is optional
                    if SHOW_STORY:
                        print(f"INFO optional recipe {recipe_line} failed.")
            else:  # no recipe, so it's a normal ending of compression
                complete_compression = True
                break  # nothing to compress
        # finish compression if the recipe asks so (and is not an uncompleted breakable).
        if recipe_line and recipe_line.islast and not (recipe_line.isbreakable and not recipe_completed):
            break
    if TIMERS:
        timer_output = get_time()

    # write the bubble
    head_comment = ''
    if not complete_compression:
        head_comment = 'Warning: incomplete compression (stopped at step {})'.format(step)
    yield from graph.bubble_repr(head_comment=head_comment, given_uid=cc_idx)

    # timers
    if TIMERS:
        now = get_time()
        timers = round(now - timer_start, 2), round(now - timer_output, 2)
        if SHOW_STORY:
            print("TIMER since start: {}s\t\toutput generation: {}s"
                  "".format(*timers))

    # compute the statistics
    compression_statistics = tuple(graph.compression_metrics())
    if cc_idx:
        compression_statistics = (('connected component', cc_idx), *compression_statistics)
    if TIMERS:
        compression_statistics = (*compression_statistics, ('time since start', timers[0]))
    # write the statistics where available
    if const.CC_STATISTIC_FILE and cc_idx:
        with open(const.CC_STATISTIC_FILE, 'a') as fd:
            csv.writer(fd).writerow(v for _, v in compression_statistics)
    if const.BUBBLE_WITH_STATISTICS:
        yield from _gen_metrics(compression_statistics)


def compress_by_cc(fname:str, recipe_files:[str]=None) -> [str]:
    """Yield bubble lines from compression of each cc found in given filename

    recipe_files -- iterable of filenames or raw recipe, or Recipe objects

    """
    if TIMERS and const.BUBBLE_WITH_STATISTICS:
        timer = get_time()
    if recipe_files:
        if SHOW_STORY:
            print('INFO recipe files:', recipe_files)
        recipe_files = tuple([recipe_files] if isinstance(recipe_files, str) else recipe_files)
        recipes = tuple(map(Recipe.from_, recipe_files))
        def recipe_for(graph:Graph, recipes=recipes) -> Recipe:
            "Return the first recipe using a node found in given graph"
            for recipe in recipes:
                if recipe.works_on(graph):
                    return recipe
    else:
        def recipe_for(*_, **__): return None  # no recipe available

    graphs = enumerate(Graph.ccs_from_file(fname), start=1)
    stats = None
    if const.PARALLEL_CC_COMPRESSION == 1:  # simple case, allowing for global stats
        for idx, graph in graphs:
            if idx > 1:  yield ''
            yield '# CONNECTED COMPONENT {}'.format(idx)
            yield from compress(graph, cc_idx=idx, recipe=recipe_for(graph))
            if const.GLOBAL_STATISTICS:
                stats = _build_global_stats(stats, graph.compression_metrics_data())
    else:  # many processes imply a more complex system
        nb_process = const.PARALLEL_CC_COMPRESSION
        if nb_process == 0:
            graphs = ((idx, gr, recipe_for(gr)) for idx, gr in graphs)
            graphs = tuple(graphs)  # RIP memory
            nb_process = len(graphs) or 1
        with ProcessPool(nb_process) as pool:
            for lines, stats_data in pool.starmap(_func_on_graph, graphs):
                yield from lines
                if const.GLOBAL_STATISTICS:
                    stats = _build_global_stats(stats, stats_data)

    # show global stats
    if stats is not None and const.BUBBLE_WITH_STATISTICS:
        yield '# TOTAL GRAPH METRICS'
        yield from _gen_metrics(Graph.compression_metrics_from_data(stats))
    if TIMERS and const.BUBBLE_WITH_STATISTICS:
        yield '# total compression time: {}'.format(round(get_time() - timer, 2))


def _func_on_graph(idx, graph, recipe):
    """Function used by multiprocessing compression of cc. Needs to be global
    to be pickled."""
    lines = (
        '# CONNECTED COMPONENT {}'.format(idx),
        *compress(graph, cc_idx=idx, recipe=recipe)
    )
    return lines, graph.compression_metrics_data()


def _build_global_stats(prev:tuple, current:tuple):
    """Return the addition of previous and current statistics values"""
    if prev is None:
        return current
    return tuple(one + two for one, two in zip(prev, current))


def _gen_metrics(metrics:[(str, float)]) -> str:
    """Yield lines describing given iterable of metric and their value."""
    metrics = tuple(metrics)
    metric_name_size = max((len(m) for m, _ in metrics))
    for metric, value in metrics:
        if isinstance(value, float): value = round(value, 2)
        yield '# {}: {}'.format(metric.ljust(metric_name_size), value)



"""Definition of a Recipe.

"""

import itertools


class RecipeError(ValueError):
    pass


class Recipe:
    RecipeError = RecipeError

    """Basically, a list of RecipeEntry"""

    def __init__(self, lines:iter):
        self._lines = tuple(RecipeEntry(*line) for line in lines)

    def __iter__(self):
        yield from self._lines

    def __len__(self):
        return len(self._lines)

    def __getitem__(self, idx):
        return self._lines[idx]

    def __str__(self):
        nb_entry = len(self._lines)
        return f"<Recipe of {nb_entry} motif{'' if nb_entry > 1 else ''}>"

    def works_on(self, graph) -> bool:
        "True if the graph has a node referenced in self."
        for entry in self:
            if f'"{entry.one_node}"' in graph.nodes:
                return True
        return False


    @staticmethod
    def from_(obj:object) -> object:
        "Build a Recipe instance from a filename, its raw data, or an iterable of lines"
        if isinstance(obj, Recipe):
            return obj
        if isinstance(obj, str):
            if '\t' in obj:
                return Recipe.from_string(obj)
            # must be a file
            return Recipe.from_file(obj)
        return Recipe.from_lines(obj)

    @staticmethod
    def from_string(string:str) -> object:
        return Recipe.from_lines(string.splitlines(False))

    @staticmethod
    def from_lines(lines:[str]) -> object:
        """Build a Recipe instance from given file"""
        def motif_from_line(line:str) -> (str, [str], [str]) or None:
            if not line or line.count('\t') != 2:
                return None  # not a Motif ; it needs to be looked by itself
            typemotif, seta, setb = line.split('\t')
            return set(typemotif.split(',')), set(seta.split(' ')), set(setb.split(' '))

        motifs = tuple(motif for motif in
                       map(motif_from_line, map(str.strip, lines)) if motif)
        return Recipe(
            (typemotifs, seta, setb) for typemotifs, seta, setb in motifs
        )

    @staticmethod
    def from_file(fname:str) -> object:
        """Build a Recipe instance from given file"""
        if fname is None: return Recipe(())
        with open(fname) as fd:
            return Recipe.from_lines(fd)


class RecipeEntry:
    RecipeError = RecipeError
    def __init__(self, typenames:set, seta:set, setb:set):
        self.typenames = frozenset(typenames)
        self.seta = frozenset(seta)
        self.setb = frozenset(setb)
        if self.isextendable and self.isbreakable:
            raise NotImplementedError("Recipe is both extendable (open/primer option)"
                                      " and breakable.")

    @property
    def isextendable(self) -> bool:
        return bool({'open', 'primer'} & self.typenames)
    @property
    def isbreakable(self) -> bool:
        return 'breakable' in self.typenames
    @property
    def islast(self) -> bool:
        return 'last' in self.typenames
    @property
    def isrequired(self) -> bool:
        return 'optional' not in self.typenames
    @property
    def one_node(self) -> str:
        """Return one node found in sets"""
        return next(iter(self.seta), None) or next(iter(self.setb))

    @property
    def covered_edges(self) -> {frozenset((str, str))}:
        "Return edges covered by the recipe"
        quoted = lambda a: '"' + a + '"'
        return frozenset(map(frozenset, itertools.product(map(quoted, self.seta), map(quoted, self.setb))))

    def __iter__(self):
        return iter((self.typenames, self.seta, self.setb))

    def __str__(self):
        return f"<RecipeEntry for {','.join(self.typenames)} with {{{','.join(self.seta)}}}×{{{','.join(self.setb)}}}>"

    def as_asp(self, is_star:bool):
        """Return ASP atoms translating the given recipe line"""
        typenames, seta, setb = self
        if min(setb) < min(seta):  # minimal element must be in seta
            setb, seta = seta, setb
        if is_star and len(seta) == 1:  # if it's a star, then single element must be in setb
            setb, seta = seta, setb
        return '\n'.join((
            ' '.join(f'newconcept(1,"{element}").' for element in seta),
            ' '.join(f'newconcept(2,"{element}").' for element in setb),
            # '|'.join(typenames) + '.'
        ))



def conversion_rate(initial_edge, final_edge, poweredge, powernode):
    """Compute conversion rate"""
    try:
        edge = initial_edge
        poweredge = final_edge + poweredge
        return (edge - poweredge) / powernode
    except ZeroDivisionError:
        return 1.

def edge_reduction(initial_edge, final_edge, poweredge, _):
    """Compute edge reduction (percentage)"""
    try:
        edge = initial_edge
        poweredge = final_edge + poweredge
        return ((edge - poweredge) / edge) * 100
    except ZeroDivisionError:
        return 100.

def compression_ratio(initial_edge, final_edge, poweredge, _):
    """Compute data compression ratio"""
    try:
        return initial_edge / (final_edge + poweredge)
    except ZeroDivisionError:
        return 1.


def compression_metrics(initial_edge, final_edge, poweredge, powernode) -> [(str, float)]:
    """Yield pairs (name, value) describing measures and their value.

    initial_edge -- number of edge in initial graph
    final_edge -- number of (non power) edge in final graph
    poweredge -- number of poweredge in final graph
    powernode -- number of powernode in final graph

    """
    payload = initial_edge, final_edge, poweredge, powernode
    for func in (conversion_rate, edge_reduction, compression_ratio):
        yield func.__name__.replace('_', ' '), func(*payload)




import math
import itertools
from collections import defaultdict

from . import asp
from . import utils
from .utils import get_time
from .motif import Motif
from .graph import Graph
from .recipe import RecipeEntry
from .constants import (TEST_INTEGRITY, SHOW_STORY, SHOW_DEBUG, KEEP_SINGLE_NODES,
                        MULTISHOT_MOTIF_SEARCH, BICLIQUE_LOWERBOUND_MAXNEI,
                        OPTIMIZE_FOR_MEMORY, CLINGO_OPTIONS)
from . import ASP_FILES

MOTIF_ASP_FILES = ASP_FILES['process-motif'], ASP_FILES['scoring_powergraph'], (ASP_FILES['block-constraint-memory'] if OPTIMIZE_FOR_MEMORY else ASP_FILES['block-constraint-cpu'])
CLIQUE_ASP_FILES = (ASP_FILES['search-clique'], *MOTIF_ASP_FILES)
FULLBICLIQUE_ASP_FILES = (ASP_FILES['search-fullbiclique'], *MOTIF_ASP_FILES)
BICLIQUE_ASP_FILES = (ASP_FILES['search-biclique'], *MOTIF_ASP_FILES)
STAR_ASP_FILES = (ASP_FILES['search-star'], *MOTIF_ASP_FILES)


class MotifSearcher:
    """A motif searcher instance provides a search over a graph
    of a particular motif.
    If the graph must change more than described by signals,
    the motif searcher instance is invalid.

    It consists in the following client API:

    - init_for_graph: called on the graph on which the search will take place.
    - search: called to search a motif in the graph.
    - on_new_compressed_motif: called when a new motif is compressed. Information on that motif is given, notably its score.
    - covered_edges: function giving the edges covered by a given motif

    And the following methods to be overwritten (if no default) by subclasses:

    - name: the name of the motif, human-readable, making it different from the others.
    - _compute_initial_upperbound: give upperbound at first step for given graph (default: number of graph edges)
    - _compute_initial_lowerbound: idem for lowerbound (default: 2)
    - _search: called to search a motif in the graph, returns atoms found by ASP solver.

    Note that a motif searcher will maintain by itself the lowerbound
    and upperbound for the motifs, by considering that a motif score is the maximal it can do.
    Computations of these bounds may rely on complex computations
    and information taken from the graph receive by init.

    """

    def __init__(self, graph:Graph):
        self.graph = graph
        self.init_for_graph(graph)
        self.__timer = None

    def init_for_graph(self, graph:Graph):
        if hasattr(self, 'compute_initial_bounds'):  # all-in-one method
            self._lowerbound, self._upperbound = self.compute_initial_bounds(graph)
        else:  # use the standard one
            self._lowerbound = self.compute_initial_lowerbound(graph)
            self._upperbound = self.compute_initial_upperbound(graph)
        if not KEEP_SINGLE_NODES and self._lowerbound in {0, 1}:
            print("WARNING lowerbound computed for {} is {}, which is an "
                  "unexpected number.".format(self.name, self._lowerbound))
        self._lowerbound = max(2, self._lowerbound)
        if SHOW_STORY and self.lowerbound > self.upperbound:
            print("INFO Search for {} will not be functional, because bounds"
                  " ({};{}) avoid any search.".format(self.name, *self.bounds))

    @property
    def lowerbound(self) -> int:  return self._lowerbound
    @property
    def upperbound(self) -> int:  return self._upperbound
    @property
    def bounds(self) -> (int, int):  return self.lowerbound, self.upperbound

    def compute_initial_upperbound(self, graph:Graph) -> int:
        return graph.nb_edge
    def compute_initial_lowerbound(self, graph:Graph) -> int:
        return 2
    def compute_new_lowerbound(self, graph:Graph, motif:Motif) -> int:
        return 2

    def on_new_found_motif(self, motif:Motif):
        """How to react when a motif is found ?"""
        if motif.ismaximal and motif.name == self.name:
            self._upperbound = min(self._upperbound, motif.score)

    def on_new_compressed_motif(self, motif:Motif):
        """How to react when a motif is compressed ?"""
        self._lowerbound = self.compute_new_lowerbound(self.graph, motif)


    def search(self, step:int, score_to_beat:int=0, recipe:RecipeEntry=None) -> [Motif]:
        """Search for motifs, better than the one to beat."""
        self.__timer = get_time()
        if recipe and not recipe.isbreakable:
            supplementary_asp_atoms = recipe.as_asp(is_star=self.name == 'star')
            lowerbound = len(recipe.seta) * len(recipe.setb)
            upperbound = self.upperbound if recipe.isextendable else lowerbound
            if SHOW_DEBUG:
                print(f'DEBUG recipe {recipe} used.')
        elif recipe and recipe.isbreakable:
            supplementary_asp_atoms = ''
            lowerbound = 2
            upperbound = self.upperbound
            if SHOW_DEBUG:
                print(f'DEBUG breakable recipe {recipe} used.')
        else:
            supplementary_asp_atoms = ''
            lowerbound = max(self.lowerbound, score_to_beat+1)
            upperbound = self.upperbound
            if SHOW_DEBUG:
                print('DEBUG no recipe available.')
        if lowerbound > upperbound:
            if SHOW_STORY:
                print("INFO No {} search because of bounds ({};{})."
                      "".format(self.name, *self.bounds))
            return  # impossible to find a motif in such conditions
        models = self._search(step, self.graph.with_recipe(recipe), lowerbound, upperbound, supplementary_asp_atoms)
        yield from (
            # the Motif is maximal, unless a recipe was biasing the search
            Motif(self.name, model, maximal=not recipe, step=step, searcher=self)
            for model in models
        )
        self.__timer = get_time() - self.__timer

    @property
    def last_search_time(self) -> float:
        return self.__timer

    def _search(self, graph:Graph, lowerbound:int, upperbound:int) -> Motif:
        raise NotImplementedError()


    def covered_edges(self, motif:Motif) -> iter:
        """Yield edges covered by given motif."""
        raise NotImplementedError()

    def _clingo_options(self):
        return CLINGO_OPTIONS[None] + ' ' + CLINGO_OPTIONS.get(self.name, '')


class BicliqueSearcher(MotifSearcher):
    """Searcher for Bicliques, including stars."""

    name = 'biclique'

    def compute_initial_bounds(self, graph:Graph) -> int:
        """
        Maximal lowerbound is the score of biggest star, or the score of the biggest intersection.
        Minimal upperbound is the maximal possible association of the most connected nodes
        """
        neis = {node: frozenset(neighbors) for node, neighbors in graph.neighbors()}
        sorted_degrees = sorted(tuple(map(len, neis.values())), reverse=True)
        try:
            upperbound = max(idx * deg for idx, deg in enumerate(sorted_degrees, start=1) if deg > 1)
        except ValueError:
            upperbound = math.inf
        upperbound = min(upperbound, graph.nb_edge)
        biggest_star = max(len(neighbors) for neighbors in neis.values())
        if BICLIQUE_LOWERBOUND_MAXNEI <= 1:
            lowerbound = biggest_star
        elif BICLIQUE_LOWERBOUND_MAXNEI == 2:
            maxnei2 = max(len(neis[a] & neis[b]) * 2 for a, b in itertools.combinations(neis.keys(), r=2))
            lowerbound = max(maxnei2, biggest_star)
        elif BICLIQUE_LOWERBOUND_MAXNEI >= 3:
            lowerbound = 0
            for level in range(2, BICLIQUE_LOWERBOUND_MAXNEI + 1):
                try:
                    max_for_level = max(len(frozenset.intersection(*(neis[s] for s in sets))) * level for sets in itertools.combinations(neis.keys(), r=level))
                except ValueError:
                    max_for_level = 0
                if max_for_level < lowerbound: break
                if lowerbound < biggest_star: break
                lowerbound = max(lowerbound, max_for_level)
        else:
            maxnei2 = max(len(neis[a] & neis[b]) * 2 for a, b in itertools.combinations(neis.keys(), r=2))
            lowerbound = max(maxnei2, biggest_star)
        return lowerbound, upperbound


    def _search(self, step:int, graph:Graph, lowerbound:int, upperbound:int, other_atoms:str='') -> iter:
        graph = ''.join(graph.as_asp(step, filter_for_bicliques=True, lowerbound=lowerbound, upperbound=upperbound))
        if SHOW_DEBUG:
            print('MXDKJX: GRAPH:', graph + '\n' + other_atoms)
        yield from asp.solve_motif_search(step, lowerbound, upperbound,
                                          options=self._clingo_options(),
                                          files=FULLBICLIQUE_ASP_FILES,
                                          graph=graph + other_atoms)

    def covered_edges(self, sets:[frozenset]) -> iter:
        """Return the edges that are covered by given sets"""
        assert len(sets) == 2
        yield from map(frozenset, itertools.product(*sets))


class NonStarBicliqueSearcher(MotifSearcher):
    """Searcher for Bicliques with at least 2 elements in each set.

    To be combined with StarSearcher to reproduce the behavior of
    BicliqueSearcher.

    """

    name = 'non-star-biclique'

    def compute_initial_bounds(self, graph:Graph) -> int:
        """
        Maximal lowerbound is the score of the biggest intersection.
        Minimal upperbound is the maximal possible association of the most connected nodes
        """
        neis = {node: frozenset(neighbors) for node, neighbors in graph.neighbors()}
        sorted_degrees = sorted(tuple(map(len, neis.values())), reverse=True)
        try:
            upperbound = max(idx * deg for idx, deg in enumerate(sorted_degrees, start=1) if deg > 1)
        except ValueError:
            upperbound = math.inf
        upperbound = min(upperbound, graph.nb_edge)
        if BICLIQUE_LOWERBOUND_MAXNEI >= 3:
            maxnei = 0
            for level in range(2, BICLIQUE_LOWERBOUND_MAXNEI + 1):
                try:
                    max_for_level = max(len(frozenset.intersection(*(neis[s] for s in sets))) * level for sets in itertools.combinations(neis.keys(), r=level))
                except ValueError:
                    max_for_level = 0
                if max_for_level < maxnei: break
                maxnei = max_for_level
        else:
            try:
                maxnei = max(len(neis[a] & neis[b]) * 2 for a, b in itertools.combinations(neis.keys(), r=2))
            except ValueError:
                maxnei = 0
        return maxnei, upperbound

    def compute_new_lowerbound(self, graph:Graph, motif:Motif) -> int:
        return 4  # At least 2 elements in each set

    def _search(self, step:int, graph:Graph, lowerbound:int, upperbound:int, other_atoms:str='') -> iter:
        graph = ''.join(graph.as_asp(step, filter_for_bicliques=True, lowerbound=lowerbound, upperbound=upperbound))
        if SHOW_DEBUG:
            print('UHJGMR: GRAPH:', graph + '\n' + other_atoms)
        yield from asp.solve_motif_search(step, lowerbound, upperbound,
                                          options=self._clingo_options(),
                                          files=BICLIQUE_ASP_FILES,
                                          graph=graph + other_atoms)

    def covered_edges(self, sets:[frozenset]) -> iter:
        """Return the edges that are covered by given sets"""
        assert len(sets) == 2
        yield from map(frozenset, itertools.product(*sets))


class StarSearcher(MotifSearcher):
    """Searcher for Stars."""

    name = 'star'

    def __init__(self, graph:Graph):
        self.__star_size = max(len(neighbors) for _, neighbors in graph.neighbors())
        super().__init__(graph)


    def compute_initial_lowerbound(self, graph:Graph) -> int:
        return self.__star_size
    def compute_initial_upperbound(self, graph:Graph) -> int:
        return self.__star_size

    def _search(self, step:int, graph:Graph, lowerbound:int, upperbound:int, other_atoms:str='') -> iter:
        graph = ''.join(graph.as_asp(step, filter_for_stars=True, lowerbound=lowerbound, upperbound=upperbound))
        if SHOW_DEBUG:
            print('ABQSSN: GRAPH:', graph + '\n' + other_atoms)
        yield from asp.solve_motif_search(step, lowerbound, upperbound,
                                          options=self._clingo_options(),
                                          files=STAR_ASP_FILES,
                                          graph=graph + other_atoms)

    def covered_edges(self, sets:[frozenset]) -> iter:
        """Return the edges that are covered by given sets"""
        assert any(len(set) == 1 for set in sets)
        assert len(sets) == 2
        yield from map(frozenset, itertools.product(*sets))


class CliqueSearcher(MotifSearcher):
    """Searcher for Cliques."""

    name = 'clique'

    def compute_initial_bounds(self, graph:Graph) -> ():
        """Return lowerbound and upperbound for initial problems.

        If any node with clustering coefficient equals to 1 exists,
        the number of neighbors gives a lowerbound.

        Use utils.maximal_clique_size function on all nodes of graph to
        compute the biggest clique possible theoretically.

        """
        node_to_edge = lambda n: (n * (n - 1)) // 2
        min_clique_size, max_clique_size = 0, 0
        for node, neighbors, nb_edge in graph.neighbors(increasing_degree=True, nb_edges_between_neighbors=True):
            node_clique_size = utils.maximal_clique_size(nb_edge)
            max_clique_size = max(max_clique_size, node_clique_size)
            if nb_edge == node_to_edge(len(neighbors)):  # clustering coefficient is 1
                min_clique_size = max(min_clique_size, len(neighbors))
            # if clique_size > len(neighbors):  # TODO: test and prove useful that optimization
                # break  # we can't found better since it's sorted
        return max(3, node_to_edge(min_clique_size)), node_to_edge(max_clique_size)

    def compute_new_lowerbound(self, graph:Graph, motif:Motif) -> int:
        return 3  # At least 3 elements in a clique


    def _search(self, step:int, graph:Graph, lowerbound:int, upperbound:int, other_atoms:str='') -> iter:
        graph = ''.join(graph.as_asp(step, filter_for_cliques=True, lowerbound=lowerbound, upperbound=upperbound))
        if SHOW_DEBUG:
            print('OKAPOD: GRAPH:', graph + '\n' + other_atoms)
        yield from asp.solve_motif_search(step, lowerbound, upperbound,
                                          options=self._clingo_options(),
                                          files=CLIQUE_ASP_FILES,
                                          graph=graph + other_atoms)

    def covered_edges(self, sets:[frozenset]) -> iter:
        """Return the edges that are covered by given sets"""
        assert len(sets) == 1
        nodes = frozenset(next(iter(sets)))
        yield from map(frozenset, itertools.combinations(nodes, r=2))


# verify clingo options
SEARCHERS = {searcher.name: searcher for searcher in globals().values()
             if type(searcher) is type and issubclass(searcher, MotifSearcher) and searcher is not MotifSearcher}
for searcher in CLINGO_OPTIONS:
    if searcher not in SEARCHERS and searcher is not None:
        raise ValueError("CLINGO_OPTIONS got a non-valid searcher value: " + searcher)



import pkg_resources
ASP_FILES = {
    fname: pkg_resources.resource_filename(__name__, 'asp/' + fname + '.lp')
    for fname in ('search-fullbiclique', 'search-biclique', 'search-star', 'search-clique', 'process-motif', 'block-constraint-cpu', 'block-constraint-memory', 'scoring_powergraph')
}

from .routines import compress_by_cc

__version__ = '0.8.17'



import phasme
import networkx
import itertools
from collections import defaultdict
from pprint import pprint
from powergrasp import utils
from powergrasp import constants
from powergrasp import edge_filtering
from powergrasp import constants as const
from powergrasp.motif import Motif
from powergrasp.recipe import Recipe


def proper_nx_graph(graph:networkx.Graph) -> networkx.Graph:
    """Return the same graph, but with properly formatted node names."""
    def node_repr(node:str or int or float) -> str or int or float:
        """Ensure that nodes in graph are well encoded, ie that integer
        nodes are integer in graph, because ASP will understand
        and handle them as integer.
        """
        if isinstance(node, str) and node.isnumeric():
            return int(node)
        elif isinstance(node, int):
            return node
        # print('NODE REPR:', node, utils.normalized_name(node))
        return utils.normalized_name(node)  # needed to avoid collisions with constants, variable,…
    proper = type(graph)()
    repr_edges = (frozenset(map(node_repr, edge)) for edge in graph.edges)  # apply node representation
    if const.KEEP_SINGLE_NODES:
        # replace self loops by node existence notification
        for edge in repr_edges:
            if len(edge) == 1:
                proper.add_node(next(iter(edge)))
            elif len(edge) == 2:
                proper.add_edge(*edge)
            else:
                raise ValueError("invalid edge components: " + str(edge))
    else:  # ignore them
        proper.add_edges_from(edge for edge in repr_edges if len(edge) == 2)
    return proper


class Graph:
    """A graph object, exposing some data on it.

    Note that a data is computed during its first access.

    Note the uid parameter for constructor

    """
    def __init__(self, graph:str or networkx.Graph):
        if isinstance(graph, networkx.Graph):
            nxgraph = networkx.Graph(graph)
        elif isinstance(graph, str):
            nxgraph = phasme.build_graph.graph_from_file(graph)
        else:
            raise ValueError("Unexpected {}".format(graph))
        # internal graph representation
        self.__edges = map(frozenset, nxgraph.edges)
        if const.TEST_INTEGRITY:
            self.__edges = tuple(self.__edges)
            for args in self.__edges:
                if len(args) > 2:
                    print('WARNING: Weird edge: {}. It will be filtered.'.format(args))
                else:
                    assert len(args) == 2, args
        self.__edges = set(edge for edge in self.__edges if len(edge) == 2)

        # data
        self.__nodes = set(nxgraph.nodes)
        self.__uid = str(min(self.__nodes, key=str))
        self.__nb_node = len(self.__nodes)
        self.__nb_cc = networkx.number_connected_components(nxgraph)
        self._nxgraph = nxgraph if constants.KEEP_NX_GRAPH else networkx.freeze(nxgraph)
        self.__initial_number_of_edge = len(self.__edges)
        self.__active_recipe = None

        self.__hierarchy = set()  # inclusions between powernodes
        self.__powernodes = defaultdict(set)  # (step, set) -> {node in powernode}
        self.__poweredges = defaultdict(set)  # (step, set) -> (step, set)

    @staticmethod
    def ccs_from_file(filename:str) -> iter:
        """Yield graphs found in given filename. Each graph is a connected component."""
        nxgraph = proper_nx_graph(phasme.build_graph.graph_from_file(filename))
        yield from (Graph(nxgraph.subgraph(cc))
                    for cc in networkx.connected_components(nxgraph))

    def with_recipe(self, recipe:Recipe or None) -> object:
        """Change the active recipe, return self.

        The recipe will alter the ASP representation of the Graph by operating
        filtering over available edges.

        """
        self.__active_recipe = recipe
        return self

    # metrics related access
    @property
    def number_of_poweredge(self) -> int:
        return sum(1 for _ in itertools.chain.from_iterable(self.__poweredges.values()))
    @property
    def number_of_simple_edge(self) -> int:
        return len(self.__edges)
    @property
    def number_of_powernode(self) -> int:
        return len(self.__powernodes)
    @property
    def initial_number_of_edge(self) -> int:
        return self.__initial_number_of_edge

    def compression_metrics(self) -> [(str, float)]:
        """Yield the metrics of compression from self data"""
        yield from Graph.compression_metrics_from_data(self.compression_metrics_data())

    @property
    def nodes(self) -> frozenset:
        return frozenset(self.__nodes)

    @staticmethod
    def compression_metrics_from_data(data:(int, int, int, int)) -> [(str, float)]:
        """Yield the metrics of compression from the input data"""
        from .metrics import compression_metrics
        yield from compression_metrics(*data)

    def compression_metrics_data(self) -> (int, int, int, int):
        """Return the 4 values used to compute the metrics"""
        return (
            self.initial_number_of_edge,
            self.number_of_simple_edge,
            self.number_of_poweredge,
            self.number_of_powernode,
        )


    def _roots(self, include_nodes:bool=True, include_pnodes:bool=True) -> iter:
        """Yield all roots, i.e nodes and powernodes that are
        not contained by a powernode.

        """
        if include_nodes:
            contained_nodes = frozenset(itertools.chain.from_iterable(self.__powernodes.values()))
            yield from (node for node in self.__nodes if node not in contained_nodes)
        if include_pnodes:
            contained_pnodes = frozenset(itertools.chain.from_iterable((s, n) for _, __, s, n in self.__hierarchy))
            yield from (pnode for pnode in self.__powernodes if pnode not in contained_pnodes)


    def as_asp(self, step:int, powerobjects:bool=False,
               filter_for_bicliques:bool=False,
               filter_for_cliques:bool=False,
               filter_for_stars:bool=False,
               lowerbound:int=2, upperbound:int=None,
               filter_by_active_recipe:bool=True) -> [str]:
        """Yield atoms of graph's ASP string representation, in edges and hierarchy.

        powerobjects -- also include powernodes and poweredges
        filter_for_bicliques -- don't yield edges that can't be compressed as bicliques
        filter_for_cliques -- don't yield edges that can't be compressed as cliques
        filter_for_stars -- don't yield edges that can't be compressed as stars
        lowerbound -- the lowerbound for the motif to search (used by filtering)
        upperbound -- idem. Will be ignored unless filtering is enabled.
        filter_by_active_recipe -- only yield edges found in active recipe.

        """
        if upperbound is None: upperbound = len(self.__edges)
        nb_filters = sum((filter_for_bicliques, filter_for_cliques, filter_for_stars))
        assert nb_filters in {0, 1}, "Too much filters asked: " + str(nb_filters)
        assert step > 0, step
        assert (step-1) >= 0, step
        # define the edges to work on (depending of choosen filter and bounds)
        filtered_edges = self.__edges
        # filter according to recipe, or if no active recipe, according to motifs
        if self.__active_recipe and filter_by_active_recipe:
            if self.__active_recipe.isextendable:
                pass  # do not prevent any edge to be covered
            else:  # not extendable, so only use the available edges
                filtered_edges = self.__edges & self.__active_recipe.covered_edges
        elif constants.GRAPH_FILTERING:
            if filter_for_bicliques: filter = edge_filtering.for_biclique
            elif filter_for_cliques: filter = edge_filtering.for_clique
            elif filter_for_stars: filter = edge_filtering.for_star
            if filter:
                filtered_edges = filter(self._nxgraph, lowerbound, upperbound)

        # yield the wanted atoms.
        for source, target in filtered_edges:
            yield 'edge({},{}).'.format(source, target)
        for (stepa, seta), nodes in self.__powernodes.items():
            for node in nodes:
                yield 'block({},{},{}).'.format(stepa, seta, node)
        for stepa, seta, stepb, setb in self.__hierarchy:
            yield 'include_block({},{},{},{}).'.format(stepa, seta, stepb, setb)
        if powerobjects:
            raise NotImplementedError()


    def compress(self, motif:Motif):
        if const.SHOW_DEBUG:
            pprint(dict(self.__powernodes))
        # Add the poweredges
        for source, target in motif.new_poweredge:
            if const.SHOW_MOTIF_HANDLING: print('\tP-EDGE', source, target)
            self.__poweredges[source].add(target)
        # Get the powernodes as set of nodes, and build the new ones.
        if not const.COVERED_EDGES_FROM_ASP:
            powernodes = set(motif.powernodes) | set(motif.stars)  # set of nodes
        for numset, node in motif.new_powernodes:
            uid = motif.uid, numset
            if const.SHOW_MOTIF_HANDLING: print('\tP-NODE', *uid, node)
            self.__powernodes[uid].add(node)
            if not const.COVERED_EDGES_FROM_ASP: powernodes.add(uid)

        # graph edges reduction and monitoring
        if const.COVERED_EDGES_FROM_ASP:
            covered = frozenset(motif.edges_covered())
        else:
            covered = frozenset(motif.edges_covered(self.__powernodes.get(uid, {uid}) for uid in powernodes))
        nb_edges_before_compress = len(self.__edges)
        if const.TEST_INTEGRITY:
            edges = frozenset(self.__edges)
            if const.SHOW_DEBUG: print('IFBTVC GRAPH:', self.__edges)
            if const.SHOW_DEBUG: print('ZEDRBM COVER:', covered)
        self.__edges -= covered
        if constants.KEEP_NX_GRAPH:
            self._nxgraph.remove_edges_from(map(tuple, covered))
        diff = nb_edges_before_compress - len(covered) != len(self.__edges)
        if const.TEST_INTEGRITY and diff:
            diff_cov = covered - edges
            diff_edg = edges - covered
            raise ValueError("{} edges yielded by {} searcher were not in the graph: {}.\n\n"
                             "{} edges in the graph were not in the {} searcher: {}."
                             "".format(len(diff_cov), motif.name, ', '.join(map(str, diff_cov)),
                                       len(diff_edg), motif.name, ', '.join(map(str, diff_edg))))
        elif diff:
            raise ValueError("Edges yielded by {} searcher were not in the graph."
                             " Rerun with TEST_INTEGRITY to get insights."
                             "".format(motif.name))
        if const.SHOW_MOTIF_HANDLING:
            print('\tCOVER', len(covered))

        # Now the big part: hierarchy. ASP send patch to apply on it.
        for args in motif.hierachy_added:
            if const.SHOW_MOTIF_HANDLING:
                print('\tADD HIERARCHY', args)
            self.__hierarchy.add(args)
            step_parent, num_parent, step_son, num_son = args
            # nodes in the parent block must be moved to the new block.
            # without that, the node would be at the same time in two powernodes.
            nodes = self.__powernodes[step_parent, num_parent] & self.__powernodes[step_son, num_son]
            if nodes:
                self.__powernodes[step_parent, num_parent] -= nodes
                self.__powernodes[step_son, num_son] |= nodes
                if const.SHOW_MOTIF_HANDLING:
                    print('\tMOVE NODES: {} FROM {} TO {}'.format(nodes, (step_parent, num_parent), (step_son, num_son)))

        for args in motif.hierachy_removed:
            if const.SHOW_MOTIF_HANDLING:
                print('\tDEL HIERARCHY', args)
            self.__hierarchy.remove(args)

        # Checking that the total number of nodes didn't change,
        #  i.e. there is no node placed in two powernodes,
        #  would be a great idea, but there is no way to do that since there is
        #  no way to know how many nodes are on ground (not in any powernode)
        #  without using the self.__powernodes attribute.
        # import itertools
        # nb_nodes_in_pnodes = sum(1 for _ in itertools.chain.from_iterable(self.__powernodes.values()))
        # nb_nodes_in_ground = ...
        # assert nb_nodes_in_pnodes == self.nb_node, (nb_nodes_in_pnodes, self.nb_node)
        if const.SHOW_DEBUG:
            print('PNODES:')
            pprint(dict(self.__powernodes))

        if const.TEST_INTEGRITY:
            inclusions = defaultdict(set)
            for stp, nsp, sts, nss in self.__hierarchy:
                inclusions[stp, nsp].add((sts, nss))
            # reversed:
            parents = defaultdict(set)
            for parent, sons in inclusions.items():
                for son in sons:
                    parents[son].add(parent)
            # detect son with multiple parents
            multiple_parents = {son for son, parents in parents.items() if len(parents) > 1}
            if multiple_parents:
                print('ERROR MULTIPLE PARENT:', len(multiple_parents))
                pprint(multiple_parents)
                exit(1)


    def compress_all(self, motifs:iter) -> int:
        """Compress all given motifs."""
        if const.SHOW_STORY or const.SHOW_DEBUG:
            print('COMPRESS…')
        for step_diff, motif in enumerate(motifs, start=0):
            motif.increase_step(step_diff)
            self.compress(motif)
        return step_diff


    @property
    def uid(self) -> str:
        return self.__uid
    @property
    def nb_edge(self) -> int:
        return len(self.__edges)
    @property
    def nb_node(self) -> int:
        return self.__nb_node
    @property
    def nb_cc(self) -> int:
        return self.__nb_cc
    def neighbors(self, increasing_degree:bool=False,
                  nb_edges_between_neighbors:bool=False) -> iter:
        """Yield pairs (node, {neighbor}).

        increasing_degree -- yield the pairs sorted
        nb_edges_between_neighbors -- yield (node, {neighbors}, N), with N the
                                      number of edges in graph with given neighbors.

        """
        def yield_data():
            for node, neighbors in self._nxgraph.adjacency():
                neighbors = tuple(neighbors)
                if nb_edges_between_neighbors:
                    yield node, neighbors, self._nxgraph.subgraph(neighbors).size()
                else:
                    yield node, neighbors
        if increasing_degree:
            data = sorted(tuple(yield_data()), key=lambda x: len(x[1]))
        else:
            data = yield_data()
        data = tuple(data)
        yield from data


    def output(self, filename:str):
        """Write in given filename the bubble representation of the graph"""
        if const.SHOW_DEBUG:
            print('\n\n###########################\n')
            print('PNODES:')
            pprint(dict(self.__powernodes))
            print('HIERAC:')
            pprint(self.__hierarchy)
            print('PEDGES:')
            pprint(dict(self.__poweredges))
            print(' EDGES:')
            pprint(self.__edges)


        with open(filename, 'w') as fd:
            for line in self.bubble_repr():
                fd.write(line + '\n')


    def bubble_repr(self, head_comment:str='', given_uid=None) -> iter:
        """Yield lines of bubble representation"""
        if head_comment:
            yield from ('# ' + line for line in head_comment.splitlines(False))
        _format_name = lambda x: format_name(format_name(None, self.uid), x)
        if const.BUBBLE_EMBEDS_CC:  # add a powernode embedding all the graph
            embedding_pnode = 'CC-{}'.format(self.uid)
            if given_uid: embedding_pnode += ' ({})'.format(given_uid)
        if const.OUTPUT_NODE_PREFIX:
            __format_name = _format_name
            _format_name = lambda x: const.OUTPUT_NODE_PREFIX + '-' + __format_name(x)
        if constants.BUBBLE_WITH_NODES:
            for node in self.__nodes:
                yield 'NODE\t{}'.format(_format_name(node))
        if constants.BUBBLE_WITH_SETS:
            for uid in self.__powernodes:
                yield 'SET\t{}\t1.0'.format(_format_name(uid))
            if const.BUBBLE_EMBEDS_CC:  # add a powernode embedding all the graph
                yield 'SET\t{}\t1.0'.format(embedding_pnode)

        for stepa, seta, stepb, setb in self.__hierarchy:
            yield 'IN\t{}\t{}'.format(_format_name((stepb, setb)), _format_name((stepa, seta)))  # contained first
        if const.BUBBLE_EMBEDS_CC:  # add a powernode embedding all the graph
            for root in self._roots():
                yield 'IN\t{}\t{}'.format(_format_name(root), embedding_pnode)  # contained first

        for (step, numset), nodes in self.__powernodes.items():
            for node in nodes:
                yield 'IN\t{}\t{}'.format(_format_name(node), _format_name((step, numset)))

        for source, targets in self.__poweredges.items():
            for target in targets:
                source, target = sorted(tuple(map(_format_name, (source, target))))
                yield 'EDGE\t{}\t{}\t{}'.format(source, target, constants.BUBBLE_POWEREDGE_FACTOR)

        if not const.BUBBLE_WITH_SIMPLE_EDGES:
            return  # do not yield the simple edges
        for source, target in self.__edges:
            source, target = sorted(tuple(map(_format_name, (source, target))))
            yield 'EDGE\t{}\t{}\t{}'.format(source, target, constants.BUBBLE_EDGE_FACTOR)


def format_name(cc:str, name:str or tuple):
    """Return string representation of powernode of given step and set nb.

    This string representation is not ASP valid,
     but can be used as name in most file formats :

        PWRN-<cc>-<step>-<num_set>

    >>> format_name(1, 'cc')
    'cc'
    >>> format_name(1, '"A"')
    'A'
    >>> format_name(1, 'A')
    'A'
    >>> format_name(42, (1, 4))
    'PWRN-42-1-4'
    >>> format_name(23, ('cc', 4))
    'PWRN-23-cc-4'
    >>> format_name(23, ('"cc"', 4))
    'PWRN-23-cc-4'

    """
    if isinstance(name, int):
        return str(name)
    elif isinstance(name, str):
        if constants.BUBBLE_SIMPLIFY_QUOTES and name[0] == '"' and name[-1] == '"':
            return name[1:-1]
        else:
            return name
    return 'PWRN-{}-{}-{}'.format(cc, *map(lambda x: format_name(None, x), name))



"""Some constants to tune the compression behavior.

Final user may want to override it by writing in 'powergrasp.cfg' file in
either JSON or INI format.

"""
import os
import ast
import json
import itertools
import configparser
import multiprocessing
from collections import ChainMap
from . import utils

# Default values
constants = {
    # Run some integrity tests at some points. May slow the compression a lot.
    'TEST_INTEGRITY': True,

    # Show main steps of the compression.
    'SHOW_STORY': True,

    # Show motifs transformation into powergraph.
    'SHOW_MOTIF_HANDLING': False,

    # Timers.
    'TIMERS': True,

    # Statistics file to write during compression.
    'STATISTIC_FILE': None,

    # Connected components statistics file to write at the end of each cc.
    'CC_STATISTIC_FILE': None,

    # Compute statistics/metrics over all connected components.
    'GLOBAL_STATISTICS': True,

    # Include some statistics in output bubble.
    'BUBBLE_WITH_STATISTICS': True,

    # Generate and save a bubble representation of the graph at each step.
    'BUBBLE_FOR_EACH_STEP': False,

    # Keep simple edges in output.
    'BUBBLE_WITH_SIMPLE_EDGES': True,

    # Prefix to add to all (power)nodes names in output.
    'OUTPUT_NODE_PREFIX': '',

    # Show full trace of the compression. Useful for debugging.
    'SHOW_DEBUG': False,

    # Recover covered edges from ASP. If falsy, will ask motif searcher to compute the edges, which may be quicker.
    'COVERED_EDGES_FROM_ASP': False,

    # Nodes and sets are optional in the output bubble.
    'BUBBLE_WITH_NODES': True,
    'BUBBLE_WITH_SETS': True,

    # Put each connected component in a dedicated powernode.
    'BUBBLE_EMBEDS_CC': False,

    # Edges in bubble are associated to a factor
    'BUBBLE_POWEREDGE_FACTOR': '1.0',
    'BUBBLE_EDGE_FACTOR': '1.0',

    # When possible, delete the quotes around identifiers in the bubble. May lead to node name collision.
    'BUBBLE_SIMPLIFY_QUOTES': True,

    # Change them according to config file
    'CONFIG_FILE': 'powergrasp.cfg',

    # Search for multiple motif in a single search. Accelerate the solving for graph with lots of equivalent motifs.
    'MULTISHOT_MOTIF_SEARCH': True,

    # Optimization on biclique lowerbound computation. Can be costly. Deactivate with 2. With value at n, up to n neighbors are considered.
    'BICLIQUE_LOWERBOUND_MAXNEI': 2,

    # Arbitrary parameters to give to clingo (note that some, like multithreading or optmode, may already be set by other options).
    'CLINGO_OPTIONS': {},

    # Number of CPU available to clingo (or a string like '2,join' or '48,compete'), or 0 for autodetect number of CPU.
    'CLINGO_MULTITHREADING': 1,

    # Two different motifs for stars and bicliques, so the work of biclique is lighter.
    'USE_STAR_MOTIF': True,

    # When a choice is given, prefer memory over CPU.
    'OPTIMIZE_FOR_MEMORY': False,

    # If set, will keep nodes connected to nothing. Else, will discard them.
    'KEEP_SINGLE_NODES': True,

    # Ignore edges dynamically determined as impossible to compress.
    'GRAPH_FILTERING': True,

    # Perform the search for motifs in different process instead of sequentially.
    'PARALLEL_MOTIF_SEARCH': False,

    # Number of processes to work on connected components. Zero to get one per cc. One to deactivate.
    'PARALLEL_CC_COMPRESSION': 1,

    # Define in which order the motifs are tested.
    'MOTIF_TYPE_ORDER': 'star,clique,non-star-biclique,biclique',

    # TODO  Detect and postpone compression of terminal tree subgraphs
    # 'TERMINAL_TREES_POSTPONING': True,
    # TODO  Detect, delete and restore bridges
    # 'BRIDGES_CUT': True,
    # TODO  Detect and if available run specialized compression routine for: trees, triangle-free graphs, cactii.
    #  Will not do anything on a graph that does not belong to those classes.
    # 'SPECIAL_CASES_DETECTION': True,
}

_derived_constants = {
    'KEEP_NX_GRAPH': lambda c: c['GRAPH_FILTERING'],
}


def make_key(key:str) -> str:
    """Return the well formed key for constants dictionary"""
    return key.upper().replace(' ', '_')

def make_value_from_ini(found_key:str, real_key:str, section,
                        config:configparser.ConfigParser):
    """Return the int, str, dict, float or tuple equivalent to value
    found for given section and key.

    """
    assert real_key in constants, f"{real_key} key found in config file is unexpected"
    assert section in config
    assert found_key in config[section], (config[section], found_key)
    default = constants[real_key]
    if isinstance(default, bool):
        return config.getboolean(section, found_key)
    elif isinstance(default, type(None)):
        try:
            return ast.literal_eval(config[section][found_key])
        except ValueError as err:  # it's not a python literal, so it's a string
            return config[section][found_key]
    elif isinstance(default, (int, float, tuple, list, dict, set, frozenset)):
        value = config[section][found_key] or '""'  # cast empty value to empty string
        try:
            return type(default)(ast.literal_eval(value))
        except ValueError as err:  # it's not a python literal, so it's a string
            return config[section][found_key]
    elif isinstance(default, str):
        return config[section][found_key]
    else:
        raise ValueError("Non-handled option type {} for field {}"
                         "".format(type(default), real_key))

def open_config_file(fname:str) -> dict:
    """Try reading file in INI, and if it do not works, try JSON"""
    try:
        # read file, take all available sections
        config = configparser.ConfigParser()
        config.read(fname)
        if config.sections():
            return {
                key: make_value_from_ini(found_key, key, section, config)
                for section in config.sections()
                for found_key, key in map(lambda k:(k, make_key(k)), config[section])
            }
        else:
            print('ERROR input config do not have any section.')
    except configparser.MissingSectionHeaderError as err:
        ini_err = err.args[0]
        try:
            with open(fname) as fd:
                return {make_key(k): v for k, v in json.load(fd).items()}
        except json.decoder.JSONDecodeError as err:
            json_err = err.args[0]
            print("ERROR input config file is not a valid json, nor a valid ini."
                  "\nINI  error: {}\nJSON error: {}".format(ini_err, json_err))


try:
    cfg = open_config_file(constants['CONFIG_FILE'])
except FileNotFoundError:
    cfg = None
if cfg:
    for field, value in cfg.items():
        if field in constants:
            constants[field] = value
        elif constants['SHOW_STORY'] or constants['SHOW_DEBUG']:
            raise ValueError("field '{}' is not a valid field for configuration."
                             "".format(field))
    if constants['SHOW_STORY'] or constants['SHOW_DEBUG']:
        print('INFO: config file {} loaded.'.format(constants['CONFIG_FILE']))
elif constants['SHOW_STORY'] or constants['SHOW_DEBUG']:
    print("INFO no config file")


def _convert_parallel_mode_option(value:str or int) -> str:
    """Return the option to give to clingo to handle given number of CPU.

    >>> _convert_parallel_mode_option(0)
    ' --parallel-mode=4'
    >>> _convert_parallel_mode_option(1)
    ''
    >>> _convert_parallel_mode_option('2')
    ' --parallel-mode=2'
    >>> _convert_parallel_mode_option(' 2')
    ' --parallel-mode=2'
    >>> _convert_parallel_mode_option('0,split')
    ' --parallel-mode=4,split'
    >>> _convert_parallel_mode_option("'0,split'")
    ' --parallel-mode=4,split'
    >>> _convert_parallel_mode_option('"0,split"')
    ' --parallel-mode=4,split'

    """
    if isinstance(value, str):
        value = value.strip(' "\'\t\n')
        if value.isnumeric():
            return str(_convert_parallel_mode_option(int(value)))
        elif value and value.startswith('0'):
            if value[1:] not in {',compete', ',split'}:
                raise ValueError("Multithreading option value is not valid: {}"
                                 "".format(value))
            nb_cpu = str(multiprocessing.cpu_count())
            return _convert_parallel_mode_option(nb_cpu + value.lstrip('0'))
        elif value:
            return ' --parallel-mode=' + value
    elif value == 0:
        return _convert_parallel_mode_option(multiprocessing.cpu_count())
    elif value > 1:
        return ' --parallel-mode=' + str(value)
    return ''


def _convert_clingo_options(value:dict or str) -> dict:
    """Return a map from motif name to clingo options."""
    # NB: can't access MotifSearchers subclass at this point, since constants
    #  are computed at import time. We therefore can't verify the keys of the dict.
    # convert string to dict
    if isinstance(value, str):
        value = {None: value}
    # validate and return the dict
    value = {None if motif is None else motif.lower().replace(' ', '-').replace('_', '-'): options
             for motif, options in value.items()}
    value.setdefault(None, '')
    return value


def _convert_motif_type_order(value:str) -> callable:
    """Return a function returning sorted searchers according to the given value.

    For instance, if input is 'worst-upperbound-first', the function returned
    will sort input searchers according to their upperbound, smallest first.

    """
    error = lambda m: ValueError("Invalid value for option MOTIF TYPE ORDER: {} ({})".format(value, m))
    value = value.lower().replace(' ', '-').replace('_', '-')
    if ',' in value:
        order = {name: idx for idx, name in enumerate(map(str.strip, value.split(',')))}
        def ordered(searchers, *, order=order) -> iter:
            """Yield given searchers according to their priority"""
            return sorted(searchers, key=lambda s: order[s.name])
    elif '-' in value:
        which, bound, where = value = value.split('-')
        greatests = {'greatest', 'biggest'}
        smallests = {'worst', 'smallest'}
        if len(value) != 3:
            raise error("{} groups instead of 3, like 'greatest-lowerbound-first'".format(len(value)))
        if which not in greatests | smallests:
            raise error("the which value must be something like worst or greatest".format(len(value)))
        reverse = (where == 'first') != (which in smallests)
        if bound == 'lowerbound':
            key = lambda s: s.lowerbound
        elif bound == 'upperbound':
            key = lambda s: s.upperbound
        else:
            raise error("the bound value must be upperbound or lowerbound".format(len(value)))
        def ordered(searchers, *, reverse=reverse, key=key) -> iter:
            return sorted(searchers, reverse=reverse, key=key)
    elif value == 'random':  # funny, but not really useful
        def ordered(searchers) -> iter:
            import random
            searchers = list(searchers)
            random.shuffle(searchers)
            return searchers
    else:
        raise error("not a list of elements or groups, like 'star,clique,non-star-biclique,biclique' or 'greatest-lowerbound-first'")
    return ordered

def _convert_erased_file(fname:str) -> str:
    """A type to give to convertion, where value is a file to erase."""
    if fname is not None:
        with open(fname, 'w') as fd:
            pass  # just erase it
    return fname  # conserve it untouched


# Apply the value convertion, if any.
_CONVERTIONS = {
    'CLINGO_MULTITHREADING': _convert_parallel_mode_option,
    'BICLIQUE_LOWERBOUND_MAXNEI': int,
    'CLINGO_OPTIONS': _convert_clingo_options,
    'MOTIF_TYPE_ORDER': _convert_motif_type_order,
    'CC_STATISTIC_FILE': _convert_erased_file,
    'COMPRESSION_STATISTIC_FILE': _convert_erased_file,
    'STATISTIC_FILE': _convert_erased_file,
}
constants = {f: _CONVERTIONS.get(f, lambda x:x)(v) for f, v in constants.items()}
# add the derived ones
constants.update({
    field: make_value(constants)
    for field, make_value in _derived_constants.items()
})

# verifications about clingo options
if constants['CLINGO_MULTITHREADING']:
    for motif, options in constants['CLINGO_OPTIONS'].items():
        if '--parallel-mode=' in options:
            raise ValueError("Invalid option value: --parallel-mode given by both"
                             " CLINGO_OPTIONS ({}) and CLINGO_MULTITHREADING ({})."
                             "".format(options, constants['CLINGO_MULTITHREADING']))


# Put them in global access
globals().update(constants)


def print_config():
    """Print configuration in stdout"""
    # name_width = max(len(name) for name, _ in uplets)
    # value_width = max(len(repr(value)) for _, value in uplets)
    uplets = sorted(tuple(OPTIONS_CATEGORIES.items()))
    name_width = max(len(name) for _, names in uplets for name in names)
    for category, options in uplets:
        print('[{}]'.format(category))
        for option in options:
            value = repr(constants[option]).strip('"\'')
            option = option.lower().replace('_', ' ').ljust(name_width)
            print('{} = {}'.format(option, value))
        print()



OPTIONS_CATEGORIES = utils.reverse_dict({
    'TEST_INTEGRITY': 'debug',
    'SHOW_STORY': 'debug',
    'SHOW_MOTIF_HANDLING': 'debug',
    'TIMERS': 'statistics',
    'STATISTIC_FILE': 'statistics',
    'CC_STATISTIC_FILE': 'statistics',
    'GLOBAL_STATISTICS': 'statistics',
    'BUBBLE_WITH_STATISTICS': 'statistics',
    'BUBBLE_FOR_EACH_STEP': 'debug',
    'BUBBLE_WITH_SIMPLE_EDGES': 'output',
    'OUTPUT_NODE_PREFIX': 'output',
    'SHOW_DEBUG': 'debug',
    'COVERED_EDGES_FROM_ASP': 'optimization',
    'BUBBLE_WITH_NODES': 'output',
    'BUBBLE_WITH_SETS': 'output',
    'BUBBLE_POWEREDGE_FACTOR': 'output',
    'BUBBLE_EDGE_FACTOR': 'output',
    'BUBBLE_EMBEDS_CC': 'output',
    'BUBBLE_SIMPLIFY_QUOTES': 'input',
    'CONFIG_FILE': 'input',
    'MULTISHOT_MOTIF_SEARCH': 'optimization',
    'BICLIQUE_LOWERBOUND_MAXNEI': 'optimization',
    'CLINGO_OPTIONS': 'clingo',
    'CLINGO_MULTITHREADING': 'clingo',
    'USE_STAR_MOTIF': 'optimization',
    'OPTIMIZE_FOR_MEMORY': 'optimization',
    'KEEP_SINGLE_NODES': 'output',
    'KEEP_NX_GRAPH': 'optimization',
    'GRAPH_FILTERING': 'optimization',
    'PARALLEL_MOTIF_SEARCH': 'optimization',
    'PARALLEL_CC_COMPRESSION': 'optimization',
    'MOTIF_TYPE_ORDER': 'optimization',
    'CC_STATISTIC_FILE': 'statistics',

    # 'TERMINAL_TREES_POSTPONING': 'optimization',
    # 'BRIDGES_CUT': 'optimization',
    # 'SPECIAL_CASES_DETECTION': 'optimization',
})
categorized_options = frozenset(itertools.chain.from_iterable(OPTIONS_CATEGORIES.values()))
assert all(option in constants for option in categorized_options)
for constant in constants:
    if constant not in categorized_options:
        raise ValueError("Option {} has no category".format(constant))
